#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	<signal.h>
#include	<netdb.h>
#include	<fcntl.h>
#include	<time.h>
#include	<termios.h>
#include	<sys/types.h>
#include	<sys/time.h>
#include	<sys/socket.h>
#include	<netinet/in.h>

#ifdef	__linux__
#include	<linux/time.h>
#include	<linux/socket.h>
#include	<linux/in.h>
#endif

#include	"voice.h"
#include	"voice_std.h"
#include	"voice_tcp.h"
#include	"voice.conf"

extern	int		errno;

PUBLIC	int		socket_fd;
PUBLIC	int		width;
PRIVATE int		debug;
PRIVATE	int		recv_fd;
PRIVATE	int		next_bull;
PRIVATE	char	password[] = "test";
PRIVATE	char	Server_ID[255];
PRIVATE char	errwork[BUFFERLEN];
PRIVATE	char	msg_buffer[BUFFERLEN];
PRIVATE	USER_SYSTEM		user[MAX_USER];
PRIVATE	ROOM_SYSTEM		room[MAX_ROOM];
PRIVATE	BULLETIN_B		bull[MAX_BULL];
PRIVATE	struct sockaddr_in		server_addr;
PRIVATE	struct sockaddr_in		client_addr;
PRIVATE	int		mesg_len;
PRIVATE	fd_set	mask, socket_mask;
PRIVATE	FILE	*acc_log;

PRIVATE	char	*colors[] = {S_BLACK, S_BLUE, S_RED, S_MAGENTA, S_GREEN,
	S_CYAN, S_YELLOW, S_WHITE, S_INIT};

int		setup_server(char*, u_short);
int		find_free_slot(void);
int		server_write(int, char*, int);
int		server_read(int, char*, int);
void	force_logout(int);
void	main_session(void);
void	init_session(void);
void	enter_room(int);
void	in_room_check(void);
void	look_all(void);
void	look_room(void);
void	set_notice(void);
void	look_notice(void);
void	version_check(void);
void	color_control(int);
void	color_set(void);
void	do_login(void);
void	do_logout(int);
void	change_handle(void);
void	silent_handle(void);
void	change_title(void);
void	door(void);
void	room_telegram(void);
void	req_all(void);
void	process_kill(int);
void	emergency_call(int);
void	alarm_sig(int);
void	keep_alive(int);
void	rcv_keep_alive(void);
void	reset_server(void);
void	version(void);
void	r_last(void);
void	superviser(void);
void	change_mode(int);
void	super_look(void);
void	super_hungup(void);
void	super_kill(void);
void	super_close(void);
void	distribution(void);
void	distrib_all(char*);
void	distrib_all_code(int);
void	distrib_room(char*, int);
void	send_roomlog(void);
void	send_roommsg(void);
void	get_address(struct sockaddr_in, int);
char* get_client_hostname(u_char, u_char, u_char, u_char, char*);
void	log_write(char*, int);
void	reset_time(void);
void	close_all(void);
void	perform(void);
void	set_bulletin(void);
void	look_bulletin(void);
void    room_log_write(int, char*);

int setup_server(char *hostname, u_short port)
{
	int		i;
	struct hostent	*myhost;

	if((myhost = gethostbyname(hostname)) == NULL)
	{
		fprintf(stderr, "bad hostname.\n");
		return (-1);
	}

	lbzero((char*)&server_addr, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_port   = htons(port);
	lbcopy((char*)myhost->h_addr, (char*)&server_addr.sin_addr, myhost->h_length);

	if((socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		fprintf(stderr, "socket alloc failed.\n");
		return (-1);
	}

	if(bind(socket_fd, (struct sockaddr*)&server_addr, sizeof(server_addr)) == -1)
	{
		fprintf(stderr, "cannot bind.\n");
		return(-1);
	}

	if(listen(socket_fd, MAX_USER) == -1)
	{
		fprintf(stderr, "cannot listen\n");
		return -1;
	}

	printf("complete bound.\n");
	for(i=0; i<MAX_USER; i++)
		user[i].flag = FREE;
	for(i=0; i<MAX_BULL; i++)
		bull[i].flag = OFF;
	next_bull = MAX_BULL;

	return 0;
}

void alarm_sig(int signo)
{
	int		i;

	for(i=0; i<MAX_USER; i++)
		if(user[i].flag != FREE) user[i].idle_time++;

	for(i=0; i<MAX_BULL; i++)
	{
		if(bull[i].flag != OFF  )
		{
			bull[i].idle++;
			if(bull[i].idle > BULL_DES_TIME) bull[i].flag = OFF;
		}
	}


	for(i=0; i<MAX_USER; i++)
	{
		if((user[i].flag >= ROOM1) && (user[i].flag <= MAX_ROOM) && (user[i].req_idle_flag == OFF))
/*		if(user[i].flag != FREE)
*/		{
			user[i].kp_idle++; 
			if((user[i].kp_idle % 60) == 0)
			{
				keep_alive(i);
				if(++user[i].kp_count > MAX_ALIVE)
				{
					msg_buffer[1] = (char)i;
					do_logout(0);
				}
			}
		}
	}



	signal(SIGALRM, alarm_sig);
	return;
}

void keep_alive(int slot)
{
	char	work[BUFFERLEN];

	work[0] = KEEP_ALIVE;
	work[1] = 0;
	server_write(user[slot].fd, work, 2);
}

void rcv_keep_alive(void)
{
	int		slot;
	slot = (int)msg_buffer[1];
	user[slot].kp_count = 0;
}

void process_kill(int process_no)
{
	int		slot;
	char	work[BUFFERLEN];

	work[0] = EMERGENCY;
	for(slot=0; slot<MAX_USER; slot++)
	{
		if(user[slot].flag != FREE)
			server_write(user[slot].fd, work, 2);
	}
	fprintf(stderr, "Emergency stop!!....%d \n", process_no);
	debug = close(socket_fd);
#ifdef DEBUG
	if(debug == -1) log_write("VOICE WARNING --- FD close error", SYSLOG);
#endif

	exit(1);
}

void emergency_call(int process_no)
{
	int		slot;
	char	work[BUFFERLEN];

	work[0] = EMERGENCY;
	work[1] = 0;
	for(slot=0; slot<MAX_USER; slot++)
	{
		if(user[slot].flag != FREE)
			server_write(user[slot].fd, work, 2);
	}

	switch(process_no)
	{
		case SIGSEGV:
			log_write("VOICE EMERGENCY CALL --- SIGSEGV", SYSLOG);
			break;
		case SIGBUS:
			log_write("VOICE EMERGENCY CALL --- SIGBUS", SYSLOG);
			break;
		case SIGHUP:
			log_write("VOICE EMERGENCY CALL --- SIGHUP", SYSLOG);
			break;
		case SIGABRT:
			log_write("VOICE EMERGENCY CALL --- SIGABRT", SYSLOG);
			break;
		case SIGINT:
			log_write("VOICE EMERGENCY CALL	--- SIGINT", SYSLOG);
			break;
		case SIGKILL:
			log_write("VOICE EMERGENCY CALL --- SIGKILL", SYSLOG);
			break;
		case SIGPIPE:
			log_write("VOICE EMERGENCY CALL --- SIGPIPE", SYSLOG);
			break;
		case SIGTERM:
			log_write("VOICE EMERGENCY CALL --- SIGTERM", SYSLOG);
			break;
		default:
			log_write("VOICE EMERGENCY CALL --- ??????", SYSLOG);
			break;
	}
	fprintf(stderr, "Emergency stop!!....%d \n", process_no);
	debug = close(socket_fd);
#ifdef DEBUG
	if(debug == -1) log_write("VOICE WARNING --- FD close error", SYSLOG);
#endif
	exit(1);
}

void enter_room(int n_room)
{
	char		work[BUFFERLEN];
	char		*time_s;
	time_t	tm;
	int			slot;

	time(&tm);
	slot = (int)msg_buffer[1];
	if(room[n_room-1].door == OPEN)
	{
		time_s = ctime(&tm);
		cut_return(time_s);
		if((int)msg_buffer[2] == OFF)
		{
			sprintf(work, "#\a Entered room --- [%s] from %s at  %s.\n", user[slot].username,
				get_client_hostname(user[slot].address[0], user[slot].address[1],
				user[slot].address[2], user[slot].address[3], user[slot].cl_address),
				time_s);
		}
		else
		{
			sprintf(work, "#\a Entered room --- [%s] from %s used backlog at  %s.\n", user[slot].username,
				get_client_hostname(user[slot].address[0], user[slot].address[1],
				user[slot].address[2], user[slot].address[3], user[slot].cl_address),
				time_s);
		}

		distrib_room(work, n_room);
		work[0] = ENTER_OK;
		server_write(recv_fd, work, 1);
		room[n_room-1].n_menber++;
		user[slot].flag = n_room;
		if(room[n_room-1].status == FREE) room[n_room-1].status = IN_USE;

		sprintf(work, "enter%d", n_room);
		log_write(work, slot);
	}
	else
	{
		sprintf(work+1, "# Sorry, This room %d is not opened.\n", n_room);
		work[0] = CANNOT_ENTER;
		server_write(recv_fd, work, strlen(work+1)+2);
	}
}

void exit_room(void)
{
	int			n_room, slot, i;
	time_t	tm;
	char		work[BUFFERLEN];
	char		*time_s;

	time(&tm);
	slot = (int)msg_buffer[1];
	n_room = user[slot].flag;
	room[n_room-1].n_menber--;
	user[slot].flag = IN_USE;
	if(room[n_room-1].n_menber == 0)
	{
		room[n_room-1].status     = FREE;
		room[n_room-1].door       = OPEN;
		room[n_room-1].log_number = 0;
		for(i=0; i<LOG; i++) room[n_room-1].rmlog[i].mesg[0] = 0;
		sprintf(room[n_room-1].room_title, "VOICE room %d", n_room);
	}
	in_room_check();
	time_s = ctime(&tm);
	cut_return(time_s);
	sprintf(work, "#\a Exited room --- [%s]  at  %s.\n", user[slot].username,
		time_s);
	distrib_room(work, n_room);

	log_write("exit", slot);
}		

void main_session(void)
{
	fd_set	readok, exceptok;
	struct timeval	timeout;
	int			found, i;
	int			fd;
	int			client_len;

	client_len = sizeof(client_addr);
	while(1)
	{
		timeout.tv_sec  = 0;
		timeout.tv_usec = WAIT_TIME;
		readok = socket_mask;
		exceptok = socket_mask;

		found = select(width, (fd_set*)&readok, NULL, &exceptok, &timeout);
		if((found > 0) && FD_ISSET(socket_fd, &readok))
		{
			fd = accept(socket_fd, (struct sockaddr*)&client_addr, &client_len);
			if(fd == -1)
			{
				sprintf(errwork, "ACCEPT ERROR=%d", errno);
				log_write(errwork, SYSLOG);
			}
			else
			{
				recv_fd = fd;
				FD_SET(fd, &mask);
				do_login();
			}
		}

                timeout.tv_sec  = 0;
                timeout.tv_usec = WAIT_TIME;
                readok = mask;
                exceptok = mask;
		found = select(width, (fd_set*)&readok, NULL, &exceptok, &timeout);
		for(i=0; i<MAX_USER; i++)
		{
			if(user[i].fd != -1)
			{
				if((found > 0) && FD_ISSET(user[i].fd, &exceptok))
				{
					recv_fd = user[i].fd;
					mesg_len = server_read(recv_fd, msg_buffer, BUFFERLEN);
					msg_buffer[mesg_len] = 0;
					log_write(msg_buffer, SYSLOG);
				}
				else 
				if((found > 0) && FD_ISSET(user[i].fd, &readok) && (user[i].flag != FREE))
				{
					recv_fd = user[i].fd;
					mesg_len = server_read(recv_fd, msg_buffer, BUFFERLEN);
					msg_buffer[mesg_len] = 0;
					if(mesg_len != 1) switch(msg_buffer[0])
					{
					case VERSION_C:
						version_check();
						break;
					case LOGIN:
						do_login();
						break;
					case LOGOUT:
						do_logout(0);	
						break;
					case ENTER_1:
					case ENTER_2:
					case ENTER_3:
					case ENTER_4:
					case ENTER_5:
					case ENTER_6:
						reset_time();
						enter_room(msg_buffer[0]);
						send_roommsg();
						break;
					case KEEP_ALIVE:
						rcv_keep_alive();
						break;
					case R_LOG:
						send_roomlog();
						break;
					case MESSAGE:
						reset_time();
						distribution();
						break;
					case LISTEN:
						reset_time();
						look_all();
						break;
					case LOOK:
						reset_time();
						look_room();
						break;
					case S_NOTICE:
						reset_time();
						set_notice();
						break;
					case L_NOTICE:
						reset_time();
						look_notice();
						break;
					case S_BULLTN:
						reset_time();
						set_bulletin();
						break;
					case L_BULLTN:
						reset_time();
						look_bulletin();
						break;
					case C_HANDLE:
						reset_time();
						change_handle();
						break;
					case S_HANDLE:
						reset_time();
						silent_handle();
						break;
					case TITLE:
						reset_time();
						change_title();
						break;
					case COLOR_ON:
						reset_time();
						color_control(ON);
						break;
					case COLOR_OFF:
						reset_time();
						color_control(OFF);
						break;
					case COLOR_SET:
						reset_time();
						color_set();
						break;
					case DOOR:
						door();
						break;
					case PERFORM:
						perform();
						break;
					case ROOM_TELE:
						reset_time();
						room_telegram();
						break;
					case REQ_ALL:
						reset_time();
						req_all();
						break;
					case SC_VER:
						reset_time();
						version();
						break;
					case R_LAST:
						reset_time();
						r_last();
						break;
					case RESET:
						reset_server();
						break;
					case SUPERVISE:
						superviser();
						break;
					case NORMAL:
						superviser();
						break;
					case SLOOK:
						reset_time();
						super_look();	
						break;
					case SCLOSE:
						super_close();
						break;
					case SEXIT:
						reset_time();
						super_hungup();
						break;
					case SHUNG:
						super_hungup();
						break;
					case SKILL:
						super_kill();
						break;
					case EXIT:
						exit_room();
						break;
					case NOP:
						break;
					default:
						break;
					}
				}
			}
		}
	}
}

void do_login(void)
{
	char	host_buffer[BUFFERLEN];
	char	id_buffer[30];
	char	work[10];
	int		slot;
	int		i;

	i = server_read(recv_fd, msg_buffer, 4);

	if(msg_buffer[1] == REVISION)
	{
		work[0] = ACCEPT;
		work[1] = REVISION;
		work[2] = msg_buffer[1];
		work[3] = 0;
		server_write(recv_fd, work, 4);
	}
	else
	{
		work[0] = REFUSE;
		work[1] = REVISION;
		work[2] = msg_buffer[1];
		work[3] = 0;
		server_write(recv_fd, work, 4);
		return;
	}


	/* receive host information */
	i = server_read(recv_fd, host_buffer, BUFFERLEN);

	/* receive user information */
	for(i=0; i<20; i++) id_buffer[i]=0;
	i = server_read(recv_fd, id_buffer, 20);
	if(i == TCP_ERROR) return;

	if((slot = find_free_slot()) == -1)
	{
		msg_buffer[0] = CANNOT_LOGIN;
		msg_buffer[1] = 0;
		msg_buffer[2] = 0;
		server_write(recv_fd, msg_buffer, 3);
		return;
	}
	else
	{
		user[slot].color_flag = msg_buffer[2];
		user[slot].color_code = (int)msg_buffer[3];
		msg_buffer[0] = LOGIN_OK;
		msg_buffer[1] = slot;
		msg_buffer[2] = 0;
		server_write(recv_fd, msg_buffer, 3);
	}

	i = server_read(recv_fd, msg_buffer, BUFFERLEN);
	msg_buffer[i] = 0;

	get_address(client_addr, slot);
	user[slot].flag = IN_USE;
	user[slot].idle_time = 0;
	user[slot].kp_idle = 0;
	user[slot].fd = recv_fd;
	strcpy(user[slot].username, msg_buffer);
	strcpy(user[slot].id, id_buffer + 1);

	/* set host information */
	if(host_buffer[1] == ON)
		strcpy(user[slot].cl_address, host_buffer+2);
	else
		strcpy(user[slot].cl_address, "none");

	log_write("login", slot);

	/* SUPERVISER MODE */
	for(i=0; i<MAX_USER; i++)
	{
		if((user[i].sp_flag == SUPERVISE) && (user[i].flag != IN_USE))
		{
			work[0] = MESSAGE;
			work[1] = 0;
			sprintf(work+2, "#--- LOGGED IN TO VOICE  \" %s \" \n", user[slot].username);
			server_write(user[i].fd, work, strlen(work+2)+3);
		}
	}


}

void set_notice(void)
{
	int		slot;

	slot = (int)msg_buffer[1];
	strcpy(user[slot].notice, msg_buffer+2);
}

void set_bulletin(void)
{
	int		slot;
	char	work[BUFFERLEN];
	time_t	tm;
	
	slot = (int)msg_buffer[1];
	time(&tm);
	next_bull--;
	if(next_bull < 0) next_bull = MAX_BULL - 1;
printf("%d\n", next_bull);
	bull[next_bull].flag = ON;
	bull[next_bull].idle = 0;
	strcpy(bull[next_bull].username, user[slot].username);
	strcpy(bull[next_bull].id, user[slot].id);
	strcpy(bull[next_bull].stime, ctime(&tm));
	strcpy(bull[next_bull].message, msg_buffer+2);
}

void look_bulletin(void)
{
	int		slot;
	int		w_flag;
	int		i;
	int		b;
	char	work[BUFFERLEN];

	slot = (int)msg_buffer[1];
	w_flag = OFF;
	for(i=0; i<MAX_BULL; i++)
		if(bull[i].flag == ON) w_flag = ON;
	if(w_flag == OFF)
	{
		work[0] = LOOK;
		strcpy(work+1, "# Bulletin Board is EMPTY.\n");
		server_write(recv_fd, work, strlen(work+1)+2);
		work[0] = ENDOFDATA;
		server_write(recv_fd, work, 1);
		return;
	}	
	
	b = 0;
	for(i=next_bull; i<MAX_BULL; i++)
		if(bull[i].flag == ON)
		{
			work[0] = LOOK;
			sprintf(work+1, "#BOARD %2.2d: %-20s [%7s] wrote at %s", b, 
				bull[i].username, bull[i].id, bull[i].stime); 
			server_write(recv_fd, work, strlen(work+1)+2);
			sprintf(work+1, "#  %s\n", bull[i].message);
			server_write(recv_fd, work, strlen(work+1)+2);
			b++;
		}

	for(i=0; i<next_bull; i++)
		if(bull[i].flag == ON)
		{
			work[0] = LOOK;
			sprintf(work+1, "#BOARD %2.2d: %-20s [%7s] wrote at %s", b, 
				bull[i].username, bull[i].id, bull[i].stime); 
			server_write(recv_fd, work, strlen(work+1)+2);
			sprintf(work+1, "#  %s\n", bull[i].message);
			server_write(recv_fd, work, strlen(work+1)+2);
			b++;
		}
	work[0] = ENDOFDATA;
	server_write(recv_fd, work, 1);
	return;
}


void look_notice(void)
{
	int		i;
	char	work[BUFFERLEN];

	work[0] = LOOK;
	for(i=0; i<MAX_USER; i++)
	{
		if((user[i].flag >= ROOM1) && (user[i].flag <= MAX_ROOM))
		{
			sprintf(work+1, "# [%1d] %-20s : %-40s\n",
				user[i].flag, user[i].username, user[i].notice);
			server_write(recv_fd, work, strlen(work+1)+2);

			sprintf(work+1, "#     from   %-25.25s   --- (%3u.%3u.%3u.%3u)\n",
				get_client_hostname(user[i].address[0], user[i].address[1],
					user[i].address[2], user[i].address[3], user[i].cl_address),
				user[i].address[0], 
				user[i].address[1], 
				user[i].address[2], 
				user[i].address[3]);
			server_write(recv_fd, work, strlen(work+1)+2);
		}
	}	
	work[0] = ENDOFDATA;
	server_write(recv_fd, work, 1);
}

void version(void)
{
	server_write(recv_fd, Server_ID, strlen(Server_ID)+1);
}

void r_last(void)
{
	FILE	*last;
	char	work[BUFFERLEN];
	long	file_pos;
	int		i;

	last = fopen(ACCESS_LOG, "r+");
	file_pos = -3L;

	for(i=0; i<(int)msg_buffer[2];)
	{
		if((file_pos = file_seek(last, file_pos)) == -1) break;
		fgets(work+1, BUFFERLEN, last);
		if(!strncmp("logout", work+1, 6))
		{
			work[0] = R_LAST;
			server_write(recv_fd, work, strlen(work+1)+2);
			i++;
		}
		else if(!strncmp("VOICE", work+1, 5))
		{
			work[0] = R_LAST;
			server_write(recv_fd, work, strlen(work+1)+2);
			i++;
		}
	}
	fclose(last);
	work[0] = ENDOFDATA;
	server_write(recv_fd, work, 1);
}

void close_all(void)
{
	int		slot;

	for(slot=0; slot<MAX_USER; slot++)
		if(user[slot].flag != FREE)
		{
			close(user[slot].fd);
#ifdef DEBUG
	if(debug == -1) log_write("VOICE WARNING --- FD close error", SYSLOG);
#endif
		}
}

void look_all(void)
{
	char	work[BUFFERLEN];
	char	uorr[7];
	int		i;

	work[0] = LISTEN;

	for(i=0; i<MAX_ROOM; i++)
	{
		if(room[i].status == FREE) strcpy(uorr, "READY");
		else                       strcpy(uorr, "IN USE");

		sprintf(work+1, "# No. %2d:  FREE, %6s: %-20s\n", 
			i+1, uorr, room[i].room_title);
		server_write(recv_fd, work, strlen(work+1)+2);
	}
	work[0] = ENDOFDATA;
	server_write(recv_fd, work, 1);
}

void perform(void)
{
	FILE	*fin;
	char	work[BUFFERLEN];

	if((fin = fopen("/proc/loadavg", "r")) == NULL)
	{
		work[0] = LISTEN;
		sprintf(work+1, "# proc filesystem error.\n");
		server_write(recv_fd, work, strlen(work+1)+2);
		work[0] = ENDOFDATA;
		server_write(recv_fd, work, 1);
	}
	else
	{
		fgets(work+1, 100, fin);
		work[0] = LISTEN;
		fclose(fin);
		server_write(recv_fd, work, strlen(work+1)+2);
		work[0] = ENDOFDATA;
		server_write(recv_fd, work, 1);
	}
}


void req_all(void)
{
	char	work[BUFFERLEN];
	int		i;

	work[0] = REQ_ALL;
	for(i=0; i<MAX_USER; i++)
	{
		if((user[i].flag >= ROOM1) && (user[i].flag <= MAX_ROOM))
 		{
			sprintf(work+1, "# (%2d )  Room%d %-20s  [%7s] \n",
				i, user[i].flag, user[i].username, user[i].id);
			server_write(recv_fd, work, strlen(work+1)+1);
		}
	}
	work[0] = ENDOFDATA;
	server_write(recv_fd, work, 1);
	user[i].req_idle_flag = ON;
}

char* get_client_hostname(u_char a1, u_char a2, u_char a3, u_char a4, char* a5)
{
	char	addrp[4];
	static	char	h_name[80];
	struct	hostent *hname;

	addrp[0] = (char)a1;
	addrp[1] = (char)a2;
	addrp[2] = (char)a3;
	addrp[3] = (char)a4;

	if(strcmp(a5, "none"))
	{
		strcpy(h_name, a5);
	}
	else
	{
		if((hname = gethostbyaddr(addrp, 4, AF_INET)) == NULL)
			sprintf(h_name, "%3u.%3u.%3u.%3u", a1, a2, a3, a4);
		else
			strcpy(h_name, hname->h_name);
	}
	return h_name;
}

void look_room(void)
{
	char	work[BUFFERLEN];
	char	uorr[10];
	int		n_room;
	int		i;

	n_room = (int)msg_buffer[2];
	work[0] = LOOK;

	switch(room[n_room-1].door)
	{
		case FREE:
			strcpy(uorr, "READY");
			break;
		case OPEN:
			strcpy(uorr, "OPEN");
			break;
		case CLOSE:
			strcpy(uorr, "CLOSE");
			break;
		case LOCK:
			strcpy(uorr, "LOCK");
			break;
		default:
			strcpy(uorr, "PRIVATE");
			break;
		}

	sprintf(work+1, "# Room %2d status is .....\n", n_room);
	server_write(recv_fd, work, strlen(work+1)+2);
	sprintf(work+1, "# No. %2d:  FREE, %6s: %-20s\n", 
			n_room, uorr, room[n_room-1].room_title);
	server_write(recv_fd, work, strlen(work+1)+2);

	for(i=0; i<MAX_USER; i++)
	{
		if(user[i].flag == n_room)
		{
			sprintf(work+1,"# (%2d) %-16s [%8s]:SL (%-25.25s) [%2d:%2d]\n",
				i, user[i].username, user[i].id, 
				get_client_hostname(user[i].address[0], user[i].address[1],
					user[i].address[2], user[i].address[3], user[i].cl_address),
				user[i].idle_time/3600, (user[i].idle_time%3600)/60);

				server_write(recv_fd, work, strlen(work+1)+2);
		}
	}

	work[0] = ENDOFDATA;
	server_write(recv_fd, work, 1);
}

void get_address(struct sockaddr_in n_addr, int slot)
{
	lbcopy((u_char*)&n_addr.sin_addr.s_addr, (u_char*)user[slot].address, 4);
}

void version_check(void)
{
	char	work[BUFFERLEN];

	if(msg_buffer[1] == REVISION) work[0] = ACCEPT;
	else													work[0] = REFUSE;

	server_write(recv_fd, work, 1);
}


void do_logout(int c)
{
	int		slot;
	char		w[BUFFERLEN];

	w[0] = LOGOUT;
	w[1] = msg_buffer[1];
	if (c == 0) server_write(recv_fd, w, 2);

	slot = (int)msg_buffer[1];
	log_write("logout", slot);
	user[slot].flag = FREE;

	user[slot].sp_flag = NORMAL;
	user[slot].color_flag = ON;
	user[slot].color_code = C_INIT;
	user[slot].kp_count = 0;
	strcpy(user[slot].notice, "unset...");

	debug = close(user[slot].fd);
	FD_CLR(user[slot].fd, &mask);
	user[slot].fd = -1;
#ifdef DEBUG
	if(debug == -1) log_write("VOICE WARNING --- FD close error", SYSLOG);
#endif
}

int find_free_slot(void)
{
	int		slot;

	for(slot=0; slot<MAX_USER; slot++)
	{
		if(user[slot].flag == FREE)
			return	slot;
	}
	return (-1);
}

void change_handle(void)
{
	int		slot;
	int		n_room;
	char	work[BUFFERLEN];

	slot = msg_buffer[1];
	n_room = user[slot].flag;


	if((user[slot].flag >= ROOM1) && (user[slot].flag <= MAX_ROOM))
	{
		sprintf(work, "# %s[%s] changes his/her handle to \"%s\"\n",
			user[slot].username, user[slot].id, msg_buffer+2);
		distrib_room(work, n_room);
	}

	strcpy(user[slot].username, msg_buffer+2);
}

void silent_handle(void)
{
	int		slot;

	slot = msg_buffer[1];
	strcpy(user[slot].username, msg_buffer+2);
}

void color_control(int sw)
{
	int		slot;

	slot = msg_buffer[1];	
	user[slot].color_flag = sw;
}

void color_set(void)
{
	int		slot;

	slot = msg_buffer[1];
	user[slot].color_code = msg_buffer[2];
}

void change_title(void)
{
	int		slot;
	int		n_room;
	char	work[BUFFERLEN];

	slot = msg_buffer[1];
	n_room = user[slot].flag;	
	sprintf(work, "#\a [%s] changed room title : %-20s\n", 
		user[slot].username, msg_buffer+2);
	distrib_room(work, n_room);
	strcpy(room[n_room-1].room_title, msg_buffer+2);
}

void send_roommsg(void)
{
	int		i, n_room, slot;
	char		work[BUFFERLEN*2];
	char		motd_str[BUFFERLEN*2];
	FILE		*motd_in;

	slot = (int)msg_buffer[1];
	n_room = user[slot].flag - 1;

	if( (motd_in = fopen(MOTD_PATH, "rb")) == NULL) return;
	
	work[0] = MESSAGE;
	strcpy(work+1, "----- MUSICATION VOICET[o肨m点 -----\n");
	server_write(recv_fd, work, strlen(work+1)+2);
	while(fgets(motd_str, BUFFERLEN, motd_in) != NULL)
	{
		strcpy(work+1, motd_str);
		work[0] = MESSAGE;
		server_write(recv_fd, work, strlen(work+1)+2);
	}
	strcpy(work+1, "----------------------------------------------\n");
	server_write(recv_fd, work, strlen(work+1)+2);
}



void send_roomlog(void)
{
	int		i, n_room, slot;
	char	work[BUFFERLEN*2];

	slot = (int)msg_buffer[1];
	n_room = user[slot].flag - 1;

	for(i=room[n_room].log_number; i<LOG; i++)
	{
		if(room[n_room].rmlog[i].mesg[0] != 0)
		{
			work[0] = MESSAGE;
			strcpy(work+1, room[n_room].rmlog[i].mesg);
			server_write(recv_fd, work, strlen(work+1)+2);
		}
	}

  for(i=0; i<room[n_room].log_number; i++)
  {
    if(room[n_room].rmlog[i].mesg[0] != 0)
  	{
    	work[0] = MESSAGE;
    	strcpy(work+1, room[n_room].rmlog[i].mesg);
    	server_write(recv_fd, work, strlen(work+1)+2);
    }
	}

	work[0] = ENDOFDATA;
	server_write(recv_fd, work, 1);
}                                            

void distribution(void)
{
	int		slot, n_room, coln;
	char	work1[BUFFERLEN];
	char	work2[BUFFERLEN];

	slot = (int)msg_buffer[1];
	coln = user[slot].color_code;

	sprintf(work1+2, "[%s] %s", user[slot].username, msg_buffer+2);
	msg_buffer[1+strlen(msg_buffer+2)] = 0;
	sprintf(work2+2, "[%s]%s %s%s\n", user[slot].username, colors[coln],
		msg_buffer+2, colors[C_INIT]);

	work1[0] = work2[0] = MESSAGE;
	work1[1] = work2[1] = msg_buffer[1];
	n_room = user[slot].flag;

	room_log_write(n_room, work1+2);

	for(slot=0; slot<MAX_USER;slot++)
	{
		if(user[slot].flag == n_room)
			if(user[slot].color_flag == OFF)
				server_write(user[slot].fd, work1, strlen(work1+2)+3);
			else
				server_write(user[slot].fd, work2, strlen(work2+2)+3);
	}

	strcpy(room[n_room-1].rmlog[ room[n_room-1].log_number ].mesg, work1+2);
	room[n_room-1].log_number++;
	if(room[n_room-1].log_number == LOG) room[n_room-1].log_number = 0;
}

void distrib_all(char *msg)
{
	int		slot;
	char	work[BUFFERLEN];

	strcpy(work+2, msg);
	work[0] = ALL_MSG;
	work[1] = 0;

	for(slot=0; slot<MAX_USER; slot++)
	{
		if(user[slot].flag != FREE)
			server_write(user[slot].fd, work, strlen(work+2)+3);
	}
}

void distrib_all_code(int code)
{
	int		slot;
	char	work[BUFFERLEN];

	work[0] = code;
	work[1] = 0;

	for(slot=0; slot<MAX_USER; slot++)
	{
		if(user[slot].flag != FREE)
			server_write(user[slot].fd, work, 2);
	}
}

void distrib_room(char *msg, int n_room)
{
	int		slot;
	char	work[BUFFERLEN];

	strcpy(work+2, msg);
	work[0] = ROOM_MSG;
	work[1] = 0;

	room_log_write(n_room, work+2);

	for(slot=0; slot<MAX_USER; slot++)
	{
		if((user[slot].flag == n_room) && (user[slot].flag != FREE))
			server_write(user[slot].fd, work, strlen(work+2)+3);
	}
}

void door(void)
{
	int		slot, n_room;

	slot = (int)msg_buffer[2];
	n_room = user[slot].flag;
	room[n_room-1].door = msg_buffer[1];
}

void room_telegram(void)
{
	int		slot, n_room, target;
	char	work[BUFFERLEN];

	slot = msg_buffer[1];
	target = msg_buffer[2];
	n_room = user[slot].flag;

	if((user[target].flag < ROOM1) || 
		 (user[target].flag > MAX_ROOM))
	{
		work[0] = REFUSE;
		work[1] = 0;
		server_write(recv_fd, work, 2);
		return;
	}
	else
	{
		work[0] = ACCEPT;
		work[1] = 0;
		server_write(recv_fd, work, 2);
	}

	work[0] = TELEGRAM;
	work[1] = NOP;
	cut_return(msg_buffer+3);
	sprintf(work+2, "#\a TELEGRAM from %s(%2d) in room%d \n--- %s\n", 
		user[slot].username, slot, n_room, msg_buffer+3);
	server_write(user[target].fd, work, strlen(work+2)+3);
	user[slot].req_idle_flag = OFF;
}

void reset_server(void)
{
	int		slot;
	char	work[BUFFERLEN];

	distrib_all("Server is reset......please hungup your client.\n");
	work[0] = EMERGENCY;
	for(slot=0; slot<MAX_USER; slot++)
	{
		if(user[slot].flag != FREE)
		{
			server_write(user[slot].fd, work, 1);
			debug = close(user[slot].fd);
#ifdef DEBUG
	if(debug == -1) log_write("VOICE WARNING --- FD close error", SYSLOG);
#endif
		}
	}
	log_write("VOICE Server re-started...", SYSLOG);
	init_session();
}

void superviser(void)
{
	char	work[BUFFERLEN];
	int		slot;

	slot = msg_buffer[1];
	if(!strcmp(msg_buffer+2, password))
	{
		if(msg_buffer[0] == SUPERVISE)	user[slot].sp_flag = SUPERVISE;
		else if(msg_buffer[0] == NORMAL)			user[slot].sp_flag = NORMAL;
		work[0] = ACCEPT;
		server_write(user[slot].fd, work, 1);
	}
	else
	{
		work[0] = REFUSE;
		server_write(user[slot].fd, work, 1);
	}
}

void super_kill(void)
{
	char	work[BUFFERLEN];
	int		i, slot;

	slot = msg_buffer[1];
	if(user[slot].sp_flag != SUPERVISE)
	{
		work[0] = REFUSE;
		server_write(recv_fd, work, 1);
		return;
	}

	i = msg_buffer[2];
	work[0] = ACCEPT;
	server_write(recv_fd, work, 1);

	work[0] = EMERGENCY;
	server_write(user[i].fd, work, 1);
	user[i].flag = FREE;
}

void super_close(void)
{
	char	work[100];
	int		f, slot;

	slot = msg_buffer[1];
	if(user[slot].sp_flag != SUPERVISE)
	{
		work[0] = REFUSE;
		server_write(recv_fd, work, 1);
		return;
	}
	f = msg_buffer[2];
	work[0] = ACCEPT;
	server_write(recv_fd, work, 1);
	close(f);
}
	

void super_look(void)
{
	char	work[BUFFERLEN];
	int		i, slot;

	slot = msg_buffer[1];
	if(user[slot].sp_flag != SUPERVISE)
	{
		work[0] = REFUSE;
		server_write(recv_fd, work, 1);
		return;
	}

	work[0] = SLOOK;
	for(i=0; i<MAX_USER; i++)
	{
		if(user[i].flag != FREE)
		{
			sprintf(work+1, "#(%2d) %2d %2d %d %s %s (%3u.%3u.%3u.%3u) [%2d:%2d]", 
				i, 
				user[i].flag,
				user[i].sp_flag,
				user[i].fd,
				user[i].username,
				user[i].id,
				user[i].address[0], user[i].address[1],
				user[i].address[2], user[i].address[3], 
				user[i].idle_time/3600, (user[i].idle_time%3600)/60);
			server_write(recv_fd, work, strlen(work+1)+2);
		}
	}
	work[0] = ENDOFDATA;
	work[1] = 0;
	server_write(recv_fd, work, 2);
}

void super_hungup(void)
{
	char	work[BUFFERLEN];
	int		slot;

	slot = msg_buffer[1];
	if(user[slot].sp_flag != SUPERVISE)
	{
		work[0] = REFUSE;
		server_write(recv_fd, work, 1);
		return;
	}
	work[0] = ACCEPT;
	server_write(recv_fd, work, 1);
	process_kill(SHUNG);
}

void reset_time(void)
{
	int		slot;

	slot = msg_buffer[1];
	user[slot].idle_time = 0;
}

void log_write(char *status, int slot)
{
	time_t	tm;
	time(&tm);

	if((acc_log = fopen(ACCESS_LOG, "a")) == NULL)
	{
		fprintf(stderr, "cannot make log file %s\n", ACCESS_LOG);
		exit(0);
	}

	if(slot == SYSLOG)
		fprintf(acc_log, "%-48.48s  %s", status, ctime(&tm));
	else
	{
		fprintf(acc_log, "%6s:%2d:%2d %-14s [%7s] (%-14.14s) %s",
			status, slot, user[slot].fd, 
			user[slot].username, user[slot].id,
			get_client_hostname(user[slot].address[0], user[slot].address[1],
			user[slot].address[2], user[slot].address[3], user[slot].cl_address), ctime(&tm));
	}
	fclose(acc_log);
}

void init_session(void)
{
	int		i, j;
	struct  itimerval       rttimer;
	struct  itimerval       old_rttimer;

	rttimer.it_value.tv_sec     = 1;
	rttimer.it_value.tv_usec    = 0;
	rttimer.it_interval.tv_sec  = 1;
	rttimer.it_interval.tv_usec = 0;
	setitimer(ITIMER_REAL, &rttimer, &old_rttimer);
                                                                 
	sprintf(Server_ID, SERVER_ID, PATCHLEVEL);
	width = MAX_USER;
	FD_ZERO(&mask);
	FD_ZERO(&socket_mask);
	FD_SET(socket_fd, &socket_mask);

	signal(SIGHUP, emergency_call);
	signal(SIGABRT, emergency_call);
	signal(SIGINT, emergency_call);
	signal(SIGKILL, emergency_call);
	signal(SIGPIPE, emergency_call);
	signal(SIGSEGV, emergency_call);
	signal(SIGBUS, emergency_call);
	signal(SIGTERM, emergency_call);
	signal(SIGALRM, alarm_sig);


	for(i=0; i<MAX_ROOM; i++)
	{
		sprintf(room[i].room_title, "VOICE room %d", i+1);
		room[i].status = FREE;
		room[i].door = OPEN;
		room[i].n_menber = 0;
		room[i].log_number = 0;
		for(j=0; j<LOG; j++) room[i].rmlog[j].mesg[0] = 0;
	}

	for(i=0; i<MAX_USER; i++)
	{
		user[i].flag = FREE;
		user[i].sp_flag = NORMAL;
		user[i].idle_time = 0;
		user[i].color_code = C_INIT;
		user[i].color_flag = ON;
		user[i].kp_count = 0;
		user[i].kp_idle = 0;
		user[i].fd = -1;
		user[i].req_idle_flag = OFF;
		strcpy(user[i].notice, "unset...");
	}
}

int main(int argc, char **argv)
{
	char		hostname[255];

	if(gethostname(hostname, 255) == -1)
	{
		fprintf(stderr, "cannot find my hostname.\n");
		exit(99);
	}

	if(setup_server(hostname, PORT) == -1)
	{
		fprintf(stderr, "cannot setup TCP server.\n");
		exit(98);
	}

	init_session();
	log_write("VOICE Server started . . .", SYSLOG);

	main_session();
	return 255;
}

int server_write(int cfd, char *data, int len)
{
	int		rt_code;

	rt_code = check_write(cfd, data, len);
	if(rt_code != -1) return rt_code;

	force_logout(cfd);

/*
	for(i=0; i<MAX_USER; i++)
	{
		if(user[i].fd == cfd)
		{
			msg_buffer[1] = i;
			if(user[i].flag < MAX_ROOM) exit_room();
			msg_buffer[1] = i;
			do_logout(1);
			break;
		}
	}
*/
	return 0;
}

int server_read(int cfd, char *data,  int len)
{
	int		rt;

	if((rt = tcp_read(cfd, data, len)) == TCP_ERROR)
	{
		force_logout(cfd);
		msg_buffer[0] = NOP;
		return TCP_ERROR;
	}
	return rt;
}

void force_logout(int cfd)
{
	int		i;
	int		n_room;
	int		s_flag;
	int		slot;
	char	work[BUFFERLEN];

	s_flag = 0;
	log_write("Force Close", cfd);

	for(i=0; i<MAX_USER; i++)
		if((user[i].flag != FREE) && (user[i].fd == cfd))
		{
			if(user[i].flag < MAX_ROOM)
			{
				slot = i;
				s_flag = 1;
				msg_buffer[1] = (char)i;
				n_room = user[i].flag;
				exit_room();
				sprintf(work, "# This \'Room Exit\' was forced logout by Server.\n");
				distrib_room(work, n_room);

				user[i].flag = FREE;
				user[i].sp_flag = NORMAL;
				user[i].color_flag = ON;
				user[i].color_code = C_INIT;
				user[i].kp_idle = 0;
				user[i].fd = -1;
				strcpy(user[i].notice, "unset...");
			}

			debug = close(cfd);
			FD_CLR(cfd, &mask);
#ifdef DEBUG
	if(debug == -1) log_write("VOICE WARNING --- FD close error", SYSLOG);
#endif
			break;
		}
	return;
}

void in_room_check(void)
{
	int		i, j, inf;

	for(i=0; i<MAX_ROOM; i++)
	{
		inf = 0;
		for(j=0; j<MAX_USER; j++)
			if(user[j].flag == (i+1)) inf++;
		if(inf == 0)
		{
			room[i].status     = FREE;
			room[i].door       = OPEN;
			room[i].log_number = 0;
			for(j=0; j<LOG; j++) room[i].rmlog[j].mesg[0] = 0;
			sprintf(room[i].room_title, "VOICE room %d", i+1);
		}
	}
}

void room_log_write(int n_room, char *msg)
{
        FILE    *fout;
        char    logfile[255];
        char    room_no[10];
        time_t  l_tm;
        struct tm       *s_tm;

        if(ROOM_LOG_ENABLE == 0) return;
        time(&l_tm);
        s_tm = localtime(&l_tm);
        switch(n_room)
        {
                case 1: strcpy(room_no, "room1"); break;
                case 2: strcpy(room_no, "room2"); break;
                case 3: strcpy(room_no, "room3"); break;
                case 4: strcpy(room_no, "room4"); break;
                case 5: strcpy(room_no, "room5"); break;
                case 6: strcpy(room_no, "room6"); break;
                default: strcpy(room_no, "etc");  break;
        }

        sprintf(logfile, "%s/%s-%4.4d%2.2d%2.2d.txt", ROOM_LOG_PATH, room_no, s_tm->tm_year+1900, s_tm->tm_mon, s_tm->tm_mday);

        if((fout = fopen(logfile, "a")) == NULL) return;
        fprintf(fout, "%s", msg);
        fclose(fout);
}

